; Scheme 9 from Empty Space, Function Library
; By Nils M Holm, 2010
; See the LICENSE file of the S9fES package for terms of use
;
; (package <name> <exports> <body>)  ==>  unspecific
;
; PACKAGE packages the definitions in its <body> in such a
; way that they are not visible outside of its body, i.e.
; the scope of each definition is the <body> of the package.
; There must be at least one definition in <body>.
;
; <Exports> is a list of symbols. When the name X of a definition
; occurs in <exports>, a symbol with the name <name>:X will be
; made visible outside of the package. That symbol will be bound
; to the same value as X inside of the package.
;
; There may be any number of options preceding the definitions
; in the <body> of PACKAGE. All options are lists with a symbol
; beginning with a #\: in their first positions. The following
; options exist:
;
; (:MAKE-ALIASES) will create an alias named X for each exported
; symbol named <name>:X, i.e. it will allow to refer to an object
; defined in a package by the same name inside and outside of the
; package.
;
; Example:   (begin
;              (package bar (foo2 foo3)
;                (:make-aliases)
;                (define (foo-maker n x)
;                  (if (zero? n)
;                      (lambda ()
;                        x)
;                      (foo-maker
;                        (- n 1)
;                        (cons n x))))
;                (define foo2 (foo-maker 2 '()))
;                (define foo3 (foo-maker 3 '())))
;              (list (bar:foo2) (foo3)))           ==>  ((1 2) (1 2 3))

(load-from-library "for-all.scm")
(load-from-library "and-letstar.scm")
(load-from-library "letrecstar.scm")
(load-from-library "pushb.scm")

(define-syntax (package name exports . body)
  (if (not (symbol? name))
      (error "package: expected name, got" name))
  (if (not (and (list? exports)
                (for-all symbol? exports)))
      (error "package: expected list of symbols, got" exports))
  (letrec
    ((options '())
     (for-all-i
       (lambda (p x)
         (cond ((null? x)
                 #t)
               ((pair? x)
                 (and (p (car x))
                      (for-all-i p (cdr x))))
               (else
                 (p x)))))
     (decompose-definition
       (lambda (x)
         (and-let* ((_    (pair? x))
                    (_    (eq? 'define (car x)))
                    (_    (pair? (cdr x)))
                    (body (cddr x))
                    (_    (pair? body))
                    (head (cadr x))
                    (_    (or (symbol? head)
                              (and (pair? head)
                                   (for-all-i symbol? head))))
                    (_    (or (not (symbol? head))
                              (= 1 (length body)))))
           (if (symbol? head)
               (list head (car body))
               (list (car head) `(lambda ,(cdr head) ,@body))))))
     (external
       (lambda (x)
         (string->symbol
           (string-append (symbol->string name)
                          ":"
                          (symbol->string x)))))
     (make-def
       (lambda (name)
         `(define ,(external name) #f)))
     (make-set
       (lambda (name)
         `(set! ,(external name) ,name)))
     (make-alias
       (lambda (name)
         `(define ,name ,(external name))))
     (option-symbol?
       (lambda (x)
         (and (symbol? x)
              (char=? #\: (string-ref (symbol->string x) 0))))))
    (let parse-opts ((opts body))
      (cond ((null? opts)
              (error "package: missing body"))
            ((and (pair? opts)
                  (pair? (car opts))
                  (option-symbol? (caar opts)))
              (case (caar opts)
                    ((:make-aliases)
                       (push! ':make-aliases options)))
              (parse-opts (cdr opts)))
            (else
              (set! body opts))))
    (let loop ((body body)
               (defs '()))
      (if (null? body)
          (let ((names (filter (lambda (x)
                                 (or (null? exports)
                                     (memq x exports)))
                               (reverse (map car defs)))))
            `(begin ,@(map make-def names)
                    (letrec*
                      ,(reverse defs)
                      ,@(map make-set names))
                    ,@(if (memq ':make-aliases options)
                          (map make-alias names)
                          '())))
          (let ((name/val (decompose-definition (car body))))
            (cond ((not name/val)
                    (error "package: expected definition, got"
                           (car body)))
                  (else
                    (loop (cdr body) (cons name/val defs)))))))))
